#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/stat.h>
#include <xspawn.h>
#include "vio.h"
#include "dirwin.h"
#include "util.h"
#include "conf.h"
#include "xpfv.h"


struct dirwin	*Main, *Dest, *Curr;


int argtest(char **argv, int min, int max) {
	int	argc;

	for (argc=0; argv[argc]; argc++)
		;
	if (argc < min) {
		alert("Too few parameters", C_msg_time);
		return(0);
	}
	if (argc > max) {
		alert("Too many parameters", C_msg_time);
		return(0);
	}
	return(1);
}


void b_chdrv() {
	xp_showall(Main, Dest, Curr);
	dw_chdrv(Curr);
}


void b_chmod() {
	xp_showall(Main, Dest, Curr);
	xp_attr(Curr, Dest, Curr);
}


void b_copy_move(int move) {
	xp_showall(Main, Dest, Curr);
	if (Curr == Main)
		copy_move(Main, Dest, move);
	else
		copy_move(Dest, Main, move);
}


void b_delete() {
	xp_showall(Main, Dest, Curr);
	do_delete(Curr);
}


void b_edit(char **filep) {
	char	cmd[128], *editor;
	extern char	*expand();
	struct stat	st;

	if (!argtest(filep, 1, 1)) return;

	if (!C_editor) {
		alert("No editor configured", C_msg_time);
		return;
	}
	if (stat(*filep, &st) < 0) {
		alert("No such file", C_msg_time);
		return;
	}
	if (st.st_mode & S_IFDIR) {
		alert("Cannot edit directories", C_msg_time);
		return;
	}

	if (!exists(editor = expand(C_editor, NULL),
	"The configured editor does not exist"))
		return;
	sprintf(cmd, "%s %s", editor, *filep);
	expand(cmd);
	(C_swap? xsystem: system)(cmd);
}


void b_exec(char **argv) {
	static char	fullpath[PATHL], path[PATHL];

	if (!argtest(argv, 1, 32767)) return;
	if (!exists(*argv, "No such file")) return;
	if (!strchr(*argv, '\\') && !strchr(*argv, '/')) {
		getcwd(path, PATHL);
		sprintf(fullpath, "%s\\%s", path, *argv);
		*argv = fullpath;
	}
	if (xspawnv(P_WAIT, *argv, argv) < 0) {
		alert("spawn failed", C_msg_time);
	}
}


void b_goto() {
	xp_showall(Main, Dest, Curr);
	if (!dw_goto(Curr, NULL)) alert("No matches", C_msg_time);
}


void b_info() {
	xp_showall(Main, Dest, Curr);
	dw_taginfo(Curr);
}


void b_mkdir() {
	xp_showall(Main, Dest, Curr);
	do_mkdir(Curr);
}


void b_qcd() {
	xp_showall(Main, Dest, Curr);
	xp_qcd(Main, Dest, Curr);
}


void b_qff() {
	xp_showall(Main, Dest, Curr);
	xp_qff(Main, Dest, Curr);
}


void b_rmtree() {
	xp_showall(Main, Dest, Curr);
	do_rmtree(Curr);
}


void b_screen(int clear) {
	if (clear) {
		vio_color(C_dos_color);
		vio_clear();
		vio_sync();
	}
	else {
		timerc(0);
		key();
		timerc(1);
	}
}


void b_sortfiles() {
	xp_showall(Main, Dest, Curr);
	xpf_sort(Main, Dest, Curr);
}


void b_tag(int on) {
	xp_showall(Main, Dest, Curr);
	mtag(Curr, on);
}


void b_view(int type, char **filep) {
	int	mode = C_browse_mode;

	xp_showall(Main, Dest, Curr);
	if (!argtest(filep, 1, 1)) return;
	if (!exists(*filep, "File not found")) return;
	switch (type) {
		case 'A':	mode = FV_TEXT; break;
		case 'D':	mode = FV_BINARY; break;
		case 'H':	mode = FV_HEX; break;
		case 'R':	mode = FV_NROFF; break;
		case 'S':	mode = FV_SCR; break;
	}

	xpfv(*filep, &mode);
}


void b_winmove() {
	xp_showall(Main, Dest, Curr);
	move_win(Main, Dest, Curr);
}


void b_winsize() {
	xp_showall(Main, Dest, Curr);
	resize_win(Main, Dest, Curr);
}


int builtin(int argc, char *argv[]) {
	int	rc = 1;
	char	*cmd = *argv, *p, *q[2];
	char	helpfile[PATHL+1];

	for (p=cmd; *p; p++)
		if (islower(*p)) *p = toupper(*p);

	if (!strcmp(cmd, "@CHDRV")) {
		b_chdrv();
	}
	else if (!strcmp(cmd, "@CHMOD")) {
		b_chmod();
	}
	else if (!strcmp(cmd, "@CLEAR")) {
		b_screen(1);
	}
	else if (!strcmp(cmd, "@COPY")) {
		b_copy_move(0);
	}
	else if (!strcmp(cmd, "@DELETE")) {
		b_delete(0);
	}
	else if (!strcmp(cmd, "@EDIT")) {
		b_edit(&argv[1]);
	}
	else if (!strcmp(cmd, "@EXEC")) {
		b_exec(&argv[1]);
	}
	else if (!strcmp(cmd, "@GOTO")) {
		b_goto();
	}
	else if (!strcmp(cmd, "@HELP")) {
		sprintf(helpfile,
			vio_clipx>64? "%s/xphelp.xp5": "%s/xphe64.xp5",
 			C_home);
		q[0] = helpfile;
		q[1] = 0;
		b_view('R', q);
	}
	else if (!strcmp(cmd, "@HIDE")) {
		b_screen(0);
	}
	else if (!strcmp(cmd, "@INFO")) {
		b_info();
	}
	else if (!strcmp(cmd, "@MKDIR")) {
		b_mkdir();
	}
	else if (!strcmp(cmd, "@MOVE")) {
		b_copy_move(1);
	}
	else if (!strcmp(cmd, "@QCD")) {
		b_qcd();
	}
	else if (!strcmp(cmd, "@QFF")) {
		b_qff();
	}
	else if (!strcmp(cmd, "@RMTREE")) {
		b_rmtree();
	}
	else if (!strcmp(cmd, "@SORTFILES")) {
		b_sortfiles();
	}
	else if (!strcmp(cmd, "@TAG")) {
		b_tag(1);
	}
	else if (!strcmp(cmd, "@UNTAG")) {
		b_tag(0);
	}
	else if (!strncmp(cmd, "@VIEW", 5)) {
		b_view(*(cmd+5)=='.'? *(cmd+6): 0, &argv[1]);
	}
	else if (!strcmp(cmd, "@WINMOVE")) {
		b_winmove();
	}
	else if (!strcmp(cmd, "@WINSIZE")) {
		b_winsize();
	}
	else {
		rc = 0;
	}
	return(rc);
}
