#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <dos.h>
#include <dir.h>
#include "keys.h"
#include "vio.h"
#include "dirwin.h"
#include "xmem.h"
#include "util.h"
#include "conf.h"


int qcmp(struct dirent *d1, struct dirent *d2) {
	char	*p1, *p2;
	int	d;

	if ((d1->attr & FA_DIREC) && !(d2->attr & FA_DIREC)) return(-1);
	if (!(d1->attr & FA_DIREC) && (d2->attr & FA_DIREC)) return(1);
	if (C_sort_mtd == 2) {
		p1 = strrchr(d1->name, '.');
		p2 = strrchr(d2->name, '.');
		if (!p1 && p2) return(-1);
		if (p1 && !p2) return(1);
		if ((d = strcmp(p1, p2)) != 0) return(d);
	}
	return(strcmp(d1->name, d2->name));
}


struct dir *dir2mem(char *path, char *pattern) {
	struct dirent	*dlist;
	struct dir	*dir;
	struct ffblk	ff;
	char	buf[PATHL], buf2[PATHL];
	int	nfile, nomore;

	if (!path) {
		if (!getcwd(buf2, PATHL)) {
			dir = (struct dir *) xmalloc(sizeof(struct dir));
			dir->path = NULL;
			dir->nfile = 0;
			return(dir);
		}
		path = buf2;
	}
	dlist = (struct dirent *) xmalloc(sizeof(struct dirent) * MAXDENT);
	if (path[strlen(path)-1] == '\\')
		sprintf(buf, "%s*.*", path);
	else
		sprintf(buf, "%s\\*.*", path);
	nomore = findfirst(buf, &ff, 0x3f);
	nfile = 0;
	while (!nomore) {
		if ((*ff.ff_name == '.' && !ff.ff_name[1]) ||
		(ff.ff_attrib & FA_LABEL) ||
		(!fnmatch(ff.ff_name, pattern) &&
		!(ff.ff_attrib & FA_DIREC))) {
			nomore = findnext(&ff); continue;
		}
		if (!C_show_hidden &&
		(ff.ff_attrib & (FA_HIDDEN|FA_SYSTEM))) {
			nomore = findnext(&ff); continue;
		}
		if (!(ff.ff_attrib & FA_DIREC)) lowcase(ff.ff_name);
		dlist[nfile].name = xsave(ff.ff_name);
		dlist[nfile].attr = ff.ff_attrib;
		dlist[nfile].time = ff.ff_ftime;
		dlist[nfile].date = ff.ff_fdate;
		dlist[nfile].size = ff.ff_fsize;
		nfile++;
		nomore = findnext(&ff);
	}

	dlist = (struct dirent *) xrealloc(dlist, sizeof(struct dirent) * nfile);
	if (nfile && C_sort_mtd) {
		qsort(dlist, nfile, sizeof(struct dirent), qcmp);
	}

	dir = (struct dir *) xmalloc(sizeof(struct dir));
	dir->path = xsave(path);
	dir->nfile = nfile;
	dir->entries = dlist;
	return(dir);
}


char *attrstr(unsigned attr) {
	static char	buf[5];

	sprintf(buf, "%c%c%c%c",
	(attr&FA_RDONLY)? 'R': 'r',
	(attr&FA_HIDDEN)? 'H': 'h',
	(attr&FA_SYSTEM)? 'S': 's',
	(attr&FA_ARCH)? 'A': 'a');

	return(buf);
}


char *datestr(unsigned date) {
	static char	buf[9];

	sprintf(buf, "%02d-%02d-%02d",
	(date>>5) & 15,
	date & 31,
	(date >> 9) + 80);

	return(buf);
}


char *timestr(unsigned time) {
	static char	buf[6];

	sprintf(buf, "%02d:%02d",
	time >> 11,
	(time>>5) & 63);

	return(buf);
}


char *sizestr(long size, unsigned attr) {
	static char	buf[9];

	if (attr & FA_DIREC) return("   (dir)");
	sprintf(buf, "%8ld", size);

	return(buf);
}


char *namestr(char *name, int len) {
	static char	buf[256];

	sprintf(buf, "%-*.*s", len, len, name);
	if (strlen(name) > len) {
		strcpy(&buf[len-3], "...");
	}

	return(buf);
}


char *tagstr(unsigned attr) {
	static char	buf[2];

	*buf = (attr & E_TAGGED)? C_dir_tag_sym: C_dir_notag_sym;
	buf[1] = 0;
	return(buf);
}


char *dirstr(struct dirent *de, char *fmt) {
	static char	ds[81];
	int	l;

	*ds = 0;
	while (*fmt) {
		switch(*fmt) {
			case 'a': case 'A':
				strcat(ds, attrstr(de->attr)); break;
			case 'd': case 'D':
				strcat(ds, datestr(de->date)); break;
			case 'n': case 'N':
				strcat(ds, namestr(de->name, NAMELEN)); break;
			case 's': case 'S':
				strcat(ds, sizestr(de->size, de->attr)); break;
			case 't': case 'T':
				strcat(ds, timestr(de->time)); break;
			case 'x': case 'X':
				strcat(ds, tagstr(de->attr)); break;
			default:
				ds[l = strlen(ds)] = *fmt;
				ds[l+1] = 0;
		}
		fmt++;
	}

	return(ds);
}


int dir_width(char *fmt) {
	int	len = 0;

	while (*fmt) {
		switch(*fmt) {
			case 'a': case 'A':	len += 4; break;
			case 'd': case 'D':	len += 8; break;
			case 'n': case 'N':	len += NAMELEN; break;
			case 's': case 'S':	len += 8; break;
			case 't': case 'T':	len += 5; break;
			case 'x': case 'X':	len ++; break;
			default:		len++;
		}
		fmt++;
	}

	return(len);
}


void dir_destroy(struct dir *dir) {
	int	i;

	for (i=0; i<dir->nfile; i++) {
		xfree(dir->entries[i].name);
	}
	xfree(dir->entries);
	xfree(dir->path);
	xfree(dir);
}


void dw_bar(struct dirwin *dw, int on) {
	if (dw->dir->nfile < 1) return;
	if (dw->dir->entries[dw->curr].attr & E_TAGGED)
		vio_color(on? C_dir_tagbar_c: C_dir_tag_c);
	else
		vio_color(on? C_dir_bar_c: C_dir_text_c);
	vio_move(dw->x+1, dw->y+1+dw->curr-dw->top);
	vio_puts(dirstr(&dw->dir->entries[dw->curr], dw->fmt));
}


void dw_list(struct dirwin *dw) {
	int	i, j;

	if (dw->dir->nfile) {
		for (i=dw->top; i < dw->dir->nfile && i-dw->top < dw->yl-1;
		i++) {
			vio_move(dw->x+1, dw->y+i+1-dw->top);
			if (dw->dir->entries[i].attr & E_TAGGED)
				vio_color(C_dir_tag_c);
			else
				vio_color(C_dir_text_c);
			vio_puts(dirstr(&dw->dir->entries[i], dw->fmt));
		}
	}
	else {
		i = 0;
	}
	vio_color(C_dir_text_c);
	while (i-dw->top < dw->yl-1) {
		vio_move(dw->x+1, dw->y+i+1-dw->top);
		for (j=0; j<dw->xl-1; j++) vio_putc(' ');
		i++;
	}
	if (!dw->dir->nfile) {
		i = 0;
		vio_move(dw->x+1, dw->y+1);
		vio_color(C_dir_text_c);
		vio_puts("<No files>");
	}
}


void dw_refresh(struct dirwin *dw, int active) {
	char	*p;

	if (!dw->bg) return;
	vio_color(C_dir_frame_c);
	vio_frame(dw->x, dw->y, dw->x+dw->xl, dw->y+dw->yl);
	if (C_smallpath) {
		vio_move(dw->x+dw->xl/2-5, dw->y);
		vio_color(C_path_c);
		p = strrchr(dw->dir->path, '\\');
		if (*(p+1)) ++p;
		vio_putc(*dw->dir->path);
		vio_putc(':');
		vio_puts(p);
	}
	if (C_fullpath) {
		vio_move(0, vio_lines-1);
		vio_color(C_path_c);
		vio_clreol();
		pth_puts(vio_clipx, dw->dir->path);
	}
	dw_list(dw);
	if (dw->curr < dw->dir->nfile) dw_bar(dw, active);
}


void dw_show(struct dirwin *dw) {
	if (dw->bg) return;
	dw->bg = vio_protect(dw->x, dw->y, dw->x+dw->xl, dw->y+dw->yl);
	dw_refresh(dw, 0);
}


void dw_hide(struct dirwin *dw) {
	if (!dw->bg) return;
	vio_hide(dw->bg);
	dw->bg = NULL;
}


struct dirwin *dw_open(int x, int y, int len, char *path, char *pattern,
char *fmt, int master) {
	struct dirwin	*dw;

	if (x < 0 || x+dir_width(fmt)+2 >= 80) {
		x = 79-dir_width(fmt)-1;
	}
	if (y+len+1 >= 25) {
		y = 25-len-1;
	}
	dw = xmalloc(sizeof(struct dirwin));
	dw->dir = dir2mem(path, pattern);
	strcpy(dw->fmt, fmt);
	strcpy(dw->pattern, pattern);
	dw->master = master;
	dw->top = dw->curr = 0;
	dw->x = x;
	dw->y = y;
	dw->yl = len+1;
	dw->xl = dir_width(fmt)+1;
	dw->bg = NULL;
	dw_show(dw);
	dw_bar(dw, 1);
	return(dw);
}


void dw_close(struct dirwin *dw) {
	dw_hide(dw);
	dir_destroy(dw->dir);
	xfree(dw);
}


int dw_move(struct dirwin *dw, int k) {
	int	move = 1, red = 0;
	int	otop = dw->top;

	if (!dw->bg) return(0);

	switch(k) {
		case K_HOME:
			if (dw->curr < 1) break;
			dw_bar(dw, 0);
			dw->top = dw->curr = 0;
			red = (otop != dw->top);
			break;

		case K_PGUP:
			if (dw->curr < 1) break;
			dw_bar(dw, 0);
			if ((dw->curr -= (dw->yl-1)) < 0) {
				dw->curr = 0;
			}
			if ((dw->top -= (dw->yl-1)) < 0) {
				dw->top = 0;
			}
			red = (otop != dw->top);
			break;

		case K_UP:
			if (dw->curr < 1) break;
			dw_bar(dw, 0);
			if (--dw->curr < dw->top) {
				dw->top--;
				vio_revind(dw->x+1, dw->y+1,
				dw->x+dw->xl-1, dw->y+dw->yl-1);
			}
			break;

		case K_DOWN:
			if (dw->curr >= dw->dir->nfile-1) break;
			dw_bar(dw, 0);
			if (++dw->curr-dw->top >= dw->yl-1) {
				dw->top++; red = 1;
				vio_index(dw->x+1, dw->y+1,
				dw->x+dw->xl-1, dw->y+dw->yl-1);
			}
			break;

		case K_PGDN:
			dw_bar(dw, 0);
			if ((dw->curr += (dw->yl-1)) >= dw->dir->nfile) {
				dw->curr = dw->dir->nfile-1;
			}
			if ((dw->top += (dw->yl-1)) >= dw->dir->nfile) {
				if ((dw->top = dw->curr-dw->yl+2) < 0) {
					dw->top = 0;
				}
			}
			red = (otop != dw->top);
			break;

		case K_END:
			dw_bar(dw, 0);
			dw->curr = dw->dir->nfile-1;
			if ((dw->top = dw->curr-dw->yl+2) < 0) {
				dw->top = 0;
			}
			red = (otop != dw->top);
			break;

		default:
			move = 0;
	}

	if (red) dw_list(dw);
	dw_bar(dw, 1);

	return(move);
}


void tag(struct dirwin *dw) {
	if (dw->dir->entries[dw->curr].attr & E_TAGGED) {
		dw->dir->entries[dw->curr].attr &= ~E_TAGGED;
		vio_color(C_dir_text_c);
	}
	else {
		dw->dir->entries[dw->curr].attr |= E_TAGGED;
		vio_color(C_dir_tag_c);
	}

	vio_move(dw->x+1, dw->y+dw->curr+1-dw->top);
	vio_puts(dirstr(&dw->dir->entries[dw->curr], dw->fmt));
}


void mtag(struct dirwin *dw, int on) {
	int	i;
	char	pattern[NAMELEN+1];

	strcpy(pattern, "*.*");
	if (vio_prompt(-1, -1, -1, C_pmt_text_c, C_pmt_high_c, C_pmt_frame_c,
	NAMELEN, pattern, on? "Tag": "Untag") < 0)
		return;

	if (!(*pattern)) return;

	for (i=0; i<dw->dir->nfile; i++) {
		if (dw->dir->entries[i].attr & FA_DIREC) continue;
		if (fnmatch(pattern, dw->dir->entries[i].name)) {
			if (on)
				dw->dir->entries[i].attr |= E_TAGGED;
			else
				dw->dir->entries[i].attr &= ~E_TAGGED;
		}
	}
}


void mtoggle(struct dirwin *dw) {
	int	i;

	for (i=0; i<dw->dir->nfile; i++) {
		if (dw->dir->entries[i].attr & FA_DIREC) continue;
		if (dw->dir->entries[i].attr & E_TAGGED)
			dw->dir->entries[i].attr &= ~E_TAGGED;
		else
			dw->dir->entries[i].attr |= E_TAGGED;
	}
}


int dw_goto(struct dirwin *dw, char *name) {
	char	buf[NAMELEN+1];
	int	i;

	if (!name) {
		buf[0] = 0;
		if (vio_prompt(-1, -1, -1, C_pmt_text_c, C_pmt_high_c,
		C_pmt_frame_c, NAMELEN, buf, "Goto") < 0)
			return(1);
		if (!(*buf)) return(1);
		strcat(buf, (!strchr(buf, '.'))? "*.*": "*");
		name = buf;
	}

	for (i=0; i<dw->dir->nfile; i++) {
		if (fnmatch(dw->dir->entries[i].name, name)) {
			dw->curr = i;
			if ((dw->top = i-dw->yl/2+1) < 0)
				dw->top = 0;
			dw_refresh(dw, 0);
			return(1);
		}
	}
	return(0);
}


int dw_chdir(struct dirwin *dw) {
	char	thisdir[NAMELEN+1], newdir[NAMELEN+1];

	if (!(dw->dir->entries[dw->curr].attr & FA_DIREC)) return(0);
	strcpy(thisdir, strrchr(dw->dir->path, '\\')+1);
	strcpy(newdir, dw->dir->entries[dw->curr].name);
	if (chdir(newdir) < 0) {
		alert("chdir failed!", C_msg_time);
		return(1);
	}
	dir_destroy(dw->dir);
	dw->dir = dir2mem(NULL, dw->pattern);
	if (!strcmp(newdir, "..")) {
		dw_goto(dw, thisdir);
	}
	else {
		dw->curr = dw->top = 0;
	}
	dw_refresh(dw, 1);

	return(1);
}


void dw_reread(struct dirwin *dw, int stick) {
	int	top, curr, i, j;
	char	name[NAMELEN+1];
	char	path[PATHL];
	struct dir	*old;

	strcpy(path, dw->dir->path);
	top = dw->top;
	curr = dw->curr;
	strcpy(name, dw->dir->entries[curr].name);
	old = dw->dir;
	dw->dir = dir2mem(path, dw->pattern);

	if (!stick) {
		dir_destroy(old);
		dw->curr = dw->top = 0; dw_refresh(dw, 0); return;
	}

	if (dw->dir->nfile <= curr) {
		dir_destroy(old);
		if (dw_goto(dw, name)) {
			dw_refresh(dw, 0); return;
		}
		dw->curr = dw->top = 0;
		dw_refresh(dw, 0);
	}
	else {
		if (strcmp(dw->dir->entries[curr].name, name)) {
			dir_destroy(old);
			dw_refresh(dw, 0);
			return;
		}
		if (dw->dir->nfile != old->nfile) {
			dir_destroy(old);
			dw_refresh(dw, 0);
			return;
		}
		dw->curr = curr;
		dw->top = top;
		for (i=0; i<old->nfile; i++) {
			j = i;
			if (strcmp(dw->dir->entries[i].name,
			old->entries[i].name)) {
				for (j=0; j<old->nfile; j++) {
					if (!strcmp(dw->dir->entries[i].name,
					old->entries[j].name))
						break;
				}
			}
			if (j >= old->nfile) continue;
			if (old->entries[j].attr & E_TAGGED)
				dw->dir->entries[i].attr |= E_TAGGED;
		}
		dir_destroy(old);
		dw_refresh(dw, 0);
	}
}


void dw_chdrv(struct dirwin *dw) {
	struct viobuf	*vb;
	struct dfree	df;
	int	k, drive, ndrv;
	char	npath[PATHL];
	int	x, y;

	x = vio_clipx/2 - 11;
	y = vio_lines/2 - 2;
	vb = vio_protect(x, y, x+19, y+2);
	vio_color(C_mnu_frame_c);
	vio_win(x, y, x+19, y+2);
	vio_move(x+2, y+1);
	vio_color(C_mnu_text_c);
	vio_puts("Select drive: ");
	vio_color(C_mnu_bar_c);
	vio_putc(getdisk()+'A');
	vio_putc(':');
	ndrv = setdisk(getdisk());
	for (;;) {
		drive = -1;
		if ((k = key()) > 255) continue;
		if (islower(k)) k = toupper(k);
		if (isalpha(k)) {
			drive = k-'A';
			if (drive >= ndrv) {
				alert("Drive not configured", C_msg_time);
				continue;
			}
			getdfree(drive+1, &df);
			if (df.df_sclus == (unsigned) -1) {
				alert("Drive not available", C_msg_time);
				continue;
			}
			break;
		}
		else if (k == K_ESC || k == K_CR) break;
	}

	vio_hide(vb);
	if (drive < 0) return;

	setdisk(drive);
	getcwd(npath, PATHL);
	xfree(dw->dir->path);
	dw->dir->path = xsave(npath);

	dw_reread(dw, 0);
}


void dw_taginfo(struct dirwin *dw) {
	unsigned long	size = 0L;
	int	n = 0, i, x, dy;
	struct viobuf	*vb;
	char	nb[11];

	for (i=0; i<dw->dir->nfile; i++) {
		if (dw->dir->entries[i].attr & E_TAGGED) {
			n++;
			size += dw->dir->entries[i].size;
		}
	}

	if (size > 10485760L)
		dy = 6;
	else if (size > 10240L)
		dy = 5;
	else
		dy = 4;
	x = (dw->x > 49)? 49: dw->x;
	vb = vio_protect(x, 0, x+30, dy);
	vio_color(C_inf_text_c);
	vio_win(x, 0, x+30, dy);
	vio_move(x+2, 1); vio_puts("There are      tagged files");
	vio_move(x+2, 2); vio_puts("containing ...");
	vio_move(x+23, 3); vio_puts("Bytes");
	if (size > 10240L) {
		vio_move(x+23, 4); vio_puts("KBytes");
		vio_color(C_inf_high_c);
		sprintf(nb, "%10ld", size / 1024L);
		vio_move(x+12, 4);
		vio_puts(nb);
	}
	if (size > 10485760L) {
		vio_color(C_inf_text_c);
		vio_move(x+23, 5); vio_puts("MBytes");
		vio_color(C_inf_high_c);
		sprintf(nb, "%10ld", size / 1048576L);
		vio_move(x+12, 5);
		vio_puts(nb);
	}
	vio_color(C_inf_high_c);
	sprintf(nb, "%10ld", size);
	vio_move(x+12, 3);
	vio_puts(nb);
	sprintf(nb, "%4d", n);
	vio_move(x+12, 1);
	vio_puts(nb);
	key();
	vio_hide(vb);
}


void dw_setpat(struct dirwin *dw) {
	char	nw[NAMELEN+1];

	strcpy(nw, dw->pattern);
	if (vio_prompt(-1, -1, -1, C_pmt_text_c, C_pmt_high_c, C_pmt_frame_c,
	NAMELEN, nw, "DIR") < 0) return;
	strcpy(dw->pattern, nw);
	dw_reread(dw, 1);
}


int dw_action(struct dirwin *dw, int k) {
	int	action = 1, red = 0;

	if (!dw->bg) return(0);

	if (k == '.' && C_dot_tags) k = K_BKSP;

	switch(k) {
		case K_BKSP:	tag(dw);
				if (C_tag_moves) dw_move(dw, K_DOWN);
				break;
		case '+':
		case '-':	mtag(dw, k=='+'); red = 1; break;
		case '*':	mtoggle(dw); red = 1; break;
		case K_LEFT:	if (dw->dir->path[strlen(dw->dir->path)-1]
				== '\\') {
					alert("Already at top", C_msg_time);
					break;
				}
		case K_RGHT:	if (!C_arrow_nav) {
					action = 0;
					break;
				}
				if (k == K_LEFT) dw_move(dw, K_HOME);
		case K_CR:	return(dw_chdir(dw));
		case ctl('D'):	dw_chdrv(dw); break;
		case ctl('G'):	if (!dw_goto(dw, NULL))
					alert("No matches", C_msg_time);
				red = 1; break;
		case ctl('R'):	dw_reread(dw, 1); break;
		case ctl('T'):	dw_taginfo(dw); break;
		case ctl('W'):	dw_setpat(dw); break;
		default:	action = 0;
	}

	if (red) dw_refresh(dw, 0);
	dw_bar(dw, 1);

	return(action);
}
