#include <stdio.h>
#include <string.h>
#include <dos.h>
#include <dir.h>
#include <sys/stat.h>
#include "keys.h"
#include "vio.h"
#include "dirwin.h"
#include "util.h"
#include "conf.h"
#include "xmem.h"


#define CPBUFSZ	65500U

extern char		*basename();
extern struct viobuf	*info();


int do_copy(char *src, char *dest, int *go) {
	struct stat	st;
	long	size, ndone;
	struct viobuf	*vb;
	int	todir, rc = -1, pcent;
	char	msg[80], dpath[PATHL], *cpmsg;
	char	*cbuf;
	unsigned int	nread;
	FILE	*in, *out;

	cpmsg = "Copying \"%s\" (%d%%)";

	if (stat(dest, &st) < 0) {
		todir = !strcmp(dest+1, ":\\") || !strcmp(dest, "\\");
	}
	else {
		todir = ((st.st_mode & S_IFDIR) != 0);
	}

	if (todir) {
		sprintf(dpath, "%s%s%s", dest,
		(dest[strlen(dest)-1] == '\\')? "": "\\",
		src);
		dest = dpath;
	}

	if (dest[1] == ':' && !dest[2]) {
		todir = 1;
		sprintf(dpath, "%s%s", dest, src);
		dest = dpath;
	}

	stat(src, &st);
	size = st.st_size;

	sprintf(msg, cpmsg, src, 0);
	vb = info(msg, NULL);

	if ((in = fopen(src, "rb")) == NULL) {
		sprintf(msg, "Cannot open: \"%s\"", src);
		alert(msg, 0);
		goto quit;
	}

	if (!(*go)) {
		if ((out = fopen(dest, "r")) != NULL) {
			fclose(out);
			sprintf(msg, "File exists: \"%s\"", src);
			alert(msg, 0);
			switch (
			menu(0, "#Skip this file|#Overwrite it|"
			"Overwrite #All")) {
				case -1:fclose(in); rc = 1; goto quit;
				case 0: fclose(in); rc = 0; goto quit;
				case 2:	*go = 1; break;
			}
		}
	}

	if ((out = fopen(dest, "wb")) == NULL) {
		fclose(in);
		sprintf(msg, "Cannot create: \"%s\"", src);
		alert(msg, 0);
		goto quit;
	}

	cbuf = xmalloc(CPBUFSZ);

	ndone = 0;
	while ((nread = fread(cbuf, 1, CPBUFSZ, in)) > 0) {
		ndone += (long) nread;
		pcent = (int) (ndone * 100 / size);
		if (pcent > 100) pcent = 100;
		sprintf(msg, cpmsg, src, pcent);
		vio_hide(vb);
		vb = info(msg, NULL);
		if (fwrite(cbuf, 1, nread, out) != nread) {
			sprintf(msg, "File write error: \"%s\"",
			basename(dest));
			alert(msg, 0);
			xfree(cbuf);
			fclose(in);
			fclose(out);
			unlink(dest);
			rc = -1;
			goto quit;
		}
	}

	xfree(cbuf);
	fclose(in);
	fclose(out);

	if (stat(dest, &st) < 0) st.st_size = -1;
	if (st.st_size != size) {
		alert("Copy failed (size mismatch)", 0);
		rc = -1;
	}
	else {
		rc = 0;
	}

quit:
	vio_hide(vb);
	return(rc);
}


int _delete(char *file, int *go) {
	char	buf[120];
	int	attr;

	if (unlink(file) < 0) {
		attr = _chmod(file, 0, 0);
		if (attr & FA_DIREC) {
			sprintf(buf, "\"%s\" is a directory", file);
			alert(buf, C_msg_time);
			return(-1);
		}
		sprintf(buf, "#Override protection %c%c%c%c for \"%s\"|"
		"#Skip this file|Override #All protections",
		(attr & FA_RDONLY)? 'R': 'r',
		(attr & FA_HIDDEN)? 'H': 'h',
		(attr & FA_SYSTEM)? 'S': 's',
		(attr & FA_ARCH)? 'A': 'a',
		basename(file));
		if (!*go) switch(menu(0, buf)) {
			case -1:return(-1);
			case 0:	break;
			case 1:	return(-1);
			case 2:	*go = 1; break;
		}
		_chmod(file, 1, 0);
		if (unlink(file) < 0) {
			sprintf(buf, "Cannot delete: \"%s\"", basename(file));
			alert(buf, C_msg_time);
			return(-1);
		}
	}
	return(0);
}


int do_move(char *src, char *dest, int *go) {
	struct stat	st;
	struct viobuf	*vb;
	int	exists, dummy;
	char	buf[80], dpath[PATHL], *cpmsg;

	cpmsg = "Moving \"%s\"";
	vb = info(cpmsg, src);
	sprintf(dpath, "%s%s%s", dest,
	(dest[strlen(dest)-1] == '\\')? "": "\\",
	src);
	if ((exists = (stat(dpath, &st) >= 0)) != 0) {
		if (st.st_mode & S_IFDIR)
			sprintf(buf, "File exists and is a directory: \"%s\"",
			src);
		else
			sprintf(buf, "File exists: \"%s\"", src);
		alert(buf, C_msg_time);
		if (st.st_mode & S_IFDIR) {
			vio_hide(vb);
			return(-1);
		}
		switch (menu(0,
		"#Skip this file|#Overwrite file|Overwrite #All")) {
			case -1:vio_hide(vb); return(-1);
			case 0:	vio_hide(vb); return(0);
			case 1:	break;
			case 2:	*go = 1; break;
		}
	}

	if (exists) {
		dummy = 1;
		if (_delete(dpath, &dummy)) {
			vio_hide(vb);
			return(-1);
		}
	}

	if (rename(src, dpath) < 0) {
		sprintf(buf, "Cannot move: \"%s\"", src);
		alert(buf, C_msg_time);
		vio_hide(vb);
		return(-1);
	}

	vio_hide(vb);
	return(0);
}


int move_conflict(char *dest) {
	char	sp[PATHL], dp[PATHL];

	getcwd(sp, PATHL);
	if (dest[1] == ':') {
		if (chdp(dest) < 0) return(-1);
	}
	else
		chdir(dest);
	getcwd(dp, PATHL);
	if (chdp(sp) < 0) return(-1);
	if (!strcmp(sp, dp)) {
		alert("Source = Destination", C_msg_time);
		return(-1);
	}
	return(0);
}


void copy_move(struct dirwin *src, struct dirwin *dest, int move) {
	extern int	Cmd_fcomp;
	int	i, n;
	char	title[80];
	char	dpath[PATHL];
	struct stat	st;
	int	go = 0;
	int	(*doit)();

	doit = move? do_move: do_copy;

	if ((n = have_tags(src)) > 0) {
		sprintf(title, "%s %d tagged files to",
		move? "Move": "Copy", n);
	}
	else {
		if (src->dir->entries[src->curr].attr & FA_DIREC) {
			alert("Cannot copy/move directories", C_msg_time);
			return;
		}
		sprintf(title, "%s \"%s\" to", move? "Move": "Copy",
		src->dir->entries[src->curr].name);
	}

	if (dest)
		strcpy(dpath, dest->dir->path);
	else
		*dpath = 0;

	Cmd_fcomp = 1;
	if (vio_prompt(-1, -1, vio_clipx-10, C_pmt_text_c, C_pmt_high_c,
	C_pmt_frame_c, PATHL, dpath, title) < 0) return;

	if (!(*dpath)) return;

	if (dpath[i = strlen(dpath)-1] == '\\') {
		if (i && dpath[i-1] != ':') dpath[i--] = 0;
	}

	if (dpath[i] != ':' && stat(dpath, &st) < 0) {
		st.st_mode = 0;
	}
	if (n || move) {
		if (dpath[1] == ':' && !dpath[2] ||
		dpath[1] == ':' && dpath[2] == '\\' && !dpath[3] ||
		dpath[0] == '\\' && !dpath[1]) {
			/* NOTHING */
		}
		else {
			if (!(st.st_mode & S_IFDIR)) {
				alert("Destination must be a directory",
				C_msg_time);
				return;
			}
		}
	}
	else {
		if (st.st_mode & S_IFDIR || dpath[i] == ':') {
			strcat(dpath, "\\");
			strcat(dpath, src->dir->entries[src->curr].name);
		}
	}

	if (move) {
		if (move_conflict(dpath)) return;
	}

	if (!n) {
		(*doit)(src->dir->entries[src->curr].name, dpath, &go);
	}
	else {
		for (i=0; i<src->dir->nfile; i++) {
			if (src->dir->entries[i].attr & E_TAGGED) {
				if ((*doit)(src->dir->entries[i].name,
				dpath, &go)) {
					if (!menu(0, "#Abort|#Continue"))
						break;
				}
				else {
					src->dir->entries[i].attr &= ~E_TAGGED;
				}
			}
		}
	}
}


int delfile(char *file, int *go) {
	struct viobuf	*vb;
	int	rc;

	vb = info("Deleting \"%s\"", file);
	rc = _delete(file, go);
	vio_hide(vb);
	return(rc);
}


void do_delete(struct dirwin *dw) {
	int	n, i, go = 0;
	char	buf[80];

	if ((n = have_tags(dw)) > 0) {
		sprintf(buf, "#Delete %d tagged files", n);
	}
	else {
		if (dw->dir->entries[dw->curr].attr & FA_DIREC) {
			alert("Cannot delete directories", C_msg_time);
			return;
		}
		sprintf(buf, "#Delete \"%s\"",
		dw->dir->entries[dw->curr].name);
	}
	strcat(buf, "|#No, keep that file(s)");
	if (menu(0, buf)) return;

	if (!n) {
		_delete(dw->dir->entries[dw->curr].name, &go);
	}
	else {
		for (i=0; i<dw->dir->nfile; i++) {
			if (dw->dir->entries[i].attr & E_TAGGED) {
				if (delfile(dw->dir->entries[i].name, &go)) {
					if (!menu(0, "#Abort|#Continue"))
						break;
				}
			}
		}
	}
}


void do_mkdir(struct dirwin *dw) {
	char	dname[NAMELEN+1], msg[80];

	*dname = 0;
	if (vio_prompt(-1, -1, -1, C_pmt_text_c, C_pmt_high_c, C_pmt_frame_c,
	NAMELEN, dname, "Make DIR") < 0) return;
	if (!(*dname)) return;

	if (mkdir(dname) < 0) {
		sprintf(msg, "Cannot create directory: \"%s\"", dname);
		alert(msg, C_msg_time);
		return;
	}

	dw_reread(dw, 0);
	dw_goto(dw, dname);
}


int rmtree(char *path, int indx, int indy) {
	struct ffblk	fb;
	int	nomore;
	static int	ind = 0;
	static char	*indc = "/-\\|";

	if (chdir(path) < 0) return(-1);
	nomore = findfirst("*.*", &fb, 0x3f);
	while (!nomore) {
		vio_move(indx, indy);
		vio_putc(indc[ind++]);
		if (ind > 3) ind = 0;
		if (!(fb.ff_attrib & FA_LABEL) && *fb.ff_name != '.') {
			if (fb.ff_attrib & (FA_RDONLY|FA_SYSTEM)) {
				_chmod(fb.ff_name, 1, 0);
			}
			if (fb.ff_attrib & FA_DIREC) {
				if (rmtree(fb.ff_name, indx, indy) < 0)
					return(-1);
			}
			else {
				if (unlink(fb.ff_name) < 0) return(-1);
			}
		}
		nomore = findnext(&fb);
	}
	if (chdir("..") < 0) return(-1);
	if (rmdir(path) < 0) return(-1);
	return(0);
}


void do_rmtree(struct dirwin *dw) {
	char	buf[80];
	struct viobuf	*bg;

	if (!(dw->dir->entries[dw->curr].attr & FA_DIREC)) {
		alert("Not a directory", C_msg_time);
		return;
	}

	if (*dw->dir->entries[dw->curr].name == '.') {
		alert("Will not remove the parent directory", C_msg_time);
		return;
	}

	if (rmdir(dw->dir->entries[dw->curr].name) < 0) {
		sprintf(buf, "Cannot remove: \"%s\"",
		dw->dir->entries[dw->curr].name);
		alert(buf, C_msg_time);
		if (menu(0, "#Alright, leave it alone|#Kill it anyway!") != 1)
			return;
		sprintf(buf, "#Remove \"%s\" and all its subdirs!|"
		"#Last chance: keep it", dw->dir->entries[dw->curr].name);
		if (menu(0, buf) != 0) return;
	}

	bg = info("Removing \"%s\":  ", dw->dir->entries[dw->curr].name);
	rmtree(dw->dir->entries[dw->curr].name, bg->dx-2, bg->y+1);
	vio_hide(bg);
}
