#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <dos.h>
#include <dir.h>
#include <time.h>
#include "vio.h"
#include "util.h"
#include "dirwin.h"
#include "conf.h"

char	*qff();


static time_t	Keytime;
int		Curpos;


char *lowcase(char *s) {
	char	*t = s;

	while (*s) {
		if (isupper(*s)) *s = tolower(*s);
		s++;
	}
	return(t);
}


void screen_saver() {
	struct viobuf	*scr;
	char	*shift = (char *) 0x417L;

	scr = vio_protect(0, 0, 79, 24);
	vio_color(0);
	vio_clear();
	while (!vio_query() && !(*shift & 0xf))
		;
	vio_hide(scr);
	Keytime = time(NULL);
}


void idle() {
	if (C_timeout && time(NULL) - Keytime >= C_timeout)
		screen_saver();

	if (C_show_timer) show_timer();
}


int key() {
	Keytime = time(NULL);
	while (!vio_query()) idle();
	return(vio_getc());
}


int timed_key(int timeout) {
	time_t	t;

	Keytime = time(NULL);
	if (!timeout) return (key());
	++timeout;
	while (timeout--) {
		t = time(NULL);
		while (t == time(NULL)) {
			idle();
			if (vio_query()) {
				Keytime = time(NULL);
				return(vio_getc());
			}
		}
	}
	return(-1);
}


void timerc(int on) {
	static int	old;

	if (!on) {
		old = C_show_timer;
		C_show_timer = 0;
	}
	else {
		C_show_timer = old;
	}
}


char *fn_expand(char *s1) {
	static char	fn[NAMELEN+1];
	int	i, d, dot = 0;

	for (i=d=0; s1[i] != 0;) {
		if (s1[i] == '.') {
			while (d < 8) fn[d++] = ' ';
			dot = 1;
			fn[d++] = '.';
			i++;
		}
		else if (s1[i] == '*') {
			if (d < 8) {
				while (d < 8) fn[d++] = '?';
				while (s1[i] != '.' && s1[i]) i++;
			}
			else {
				if (dot) {
					while (d < NAMELEN) fn[d++] = '?';
				}
				i++;
			}
		}
		else if (d < NAMELEN) {
			if (isupper(s1[i]))
				fn[d++] = tolower(s1[i++]);
			else
				fn[d++] = s1[i++];
		}
		else
			i++;
	}

	if (!dot) for (; d<8; d++) fn[d] = '?';
	fn[d] = 0;
	if (d >= 8) fn[8] = '.';
	if (d == 8) fn[9] = 0;

	return(fn);
}


int fnmatch(char *s1, char *s2) {
	char	fn1[NAMELEN+1], *fn2;
	int	i;

	strcpy(fn1, fn_expand(s1));
	fn2 = fn_expand(s2);

	for (i=0; i<NAMELEN; i++) {
		if (fn1[i] == 0 && fn2[i] == 0) {
			return(1);
		}
		else if (fn1[i] != '?' && fn2[i] != '?') {
			if (fn1[i] != fn2[i]) return(0);
		}
	}
	return(1);
}


int have_tags(struct dirwin *dw) {
	int	i, cnt = 0;

	for (i=0; i<dw->dir->nfile; i++)
		if (dw->dir->entries[i].attr & E_TAGGED)
			cnt++;
	return(cnt);
}


int menu(int dflt, char *items) {
	return(vio_menu(-1, -1, C_mnu_text_c, C_mnu_bar_c, C_mnu_high_c,
	dflt, items));
}


struct viobuf *info(char *msg, char *name) {
	struct viobuf	*vb;
	char	buf[81], x, dx, l, y;

	sprintf(buf, msg, name);
	l = strlen(buf)+4;
	x = (vio_clipx-1)/2-(l)/2;
	y = (vio_lines-1)/2-1;
	dx = x+l-1;
	vb = vio_protect(x, y, dx, y+2);
	vio_color(C_inf_text_c);
	vio_win(x, y, dx, y+2);
	vio_move(x+2, y+1);
	vio_puts(buf);
	return(vb);
}


char *expand(char *s, struct dirwin *dw) {
	int	i, j, k;
	int	vp;
	static char	t[512], vn[128], *vv;

	for (i=j=0; s[i]; i++) {
		if (s[i] == '~') {
			for (k=0; C_home[k]; k++)
				t[j++] = C_home[k];
		}
		else if (s[i] == '$') {
			if (s[++i] == '$') {
				t[j++] = s[i];
			}
			else {
				vp = 0;
				while (isupper(s[i])) {
					vn[vp++] = s[i++];
				}
				vn[vp] = 0;
				i--;
				if ((vv = getenv(vn)) == NULL) vv = "";
				for (k=0; vv[k]; k++)
					t[j++] = vv[k];
			}
		}
		else if (s[i] == '%') {
			if (s[i+1] == 's') {
				i++;
				vv = dw->dir->entries[dw->curr].name;
				for (k=0; vv[k]; k++)
					t[j++] = vv[k];
			}
			else {
				t[j++] = s[i];
			}	
		}
		else {
			t[j++] = s[i];
		}
		if (j>128) return(s);
	}
	t[j] = 0;
	strcpy(s, t);
	return(s);
}


char *basename(char *path) {
	char	*p;

	if ((p = strrchr(path, '\\')) != NULL)
		return(p+1);
	else if ((p = strrchr(path, '/')) != NULL)
		return(p+1);
	else
		return(path);

}


char *getlabel(int id) {
	static struct ffblk	lbl;
	char	buf[10];

	sprintf(buf, "%c:\\*.*", id);
	if (findfirst(buf, &lbl, FA_LABEL) != 0)
		return("");
	else
		return(lbl.ff_name);
}


static int	Disk_error;


int not_avail() {
	Disk_error = 1;
	hardretn(-1);
	return(0);
}


void diskfree() {
	extern int	fuckup();
	struct viobuf	*bg;
	struct dfree	df;
	time_t	t;
	char	buf[81], *p, *dl;
	int	id, pc, i;
	long	total, free;

	bg = vio_protect(0, 0, 79, 24);
	vio_color(C_dir_text_c);
	vio_clear();

	time(&t);
	p = ctime(&t);
	p[strlen(p)-1] = 0;
	sprintf(buf, "Disk capacities as of %s", p);
	vio_color(C_inf_high_c);
	vio_move(0, 0); vio_clreol();
	vio_move(16, 0); vio_puts(buf);

	vio_color(C_dir_tag_c);
	vio_move(15, 3);
	vio_puts("ID  Label           Capacity   Freespace   %Free");

	harderr(not_avail);
	Disk_error = 0;

	vio_color(C_dir_text_c);
	for (id='C'; id<'Z'; id++) {
		vio_move(15, 5+id-'C'); vio_color(C_dir_tag_c);
		vio_puts("-- Checking --");
		getdfree(id-'A'+1, &df);
		vio_move(0, 5+id-'C'); vio_color(C_dir_text_c); vio_clreol();
		if (Disk_error) break;
		if ((int) df.df_sclus == -1) continue;

		dl = getlabel(id);
		total = (long) df.df_sclus * df.df_bsec * df.df_total;
		free = (long) df.df_sclus * df.df_bsec * df.df_avail;
		pc = (int) (free / (total / 100));
		sprintf(buf, "%c:  %-14.14s%10ld  %10ld   %5d",
		id, dl, total, free, pc);

		vio_move(15, 5+id-'C');
		vio_puts(buf);

		if (id == 15) {
			vio_color(C_dir_tag_c);
			vio_move(15, id+7-'C');
			vio_puts(" -- more -- ");
			key();
			vio_color(C_dir_text_c);
			for (i=5; i<25; i++) {
				vio_move(0, i); vio_clreol();
			}
		}
	}

	harderr(fuckup);

	timerc(0);
	key();
	timerc(1);
	vio_hide(bg);
}


void pth_puts(int k, char *s) {
	int	l = strlen(s);

	if (l>k) vio_puts("...");
	vio_puts(l>k? s+l-k+3: s);
}


void savecpos() {
	union REGS	regs;

	regs.x.ax = 0x300;
	regs.x.bx = 0;
	int86(0x10, &regs, &regs);
	Curpos = regs.x.dx;
	regs.x.ax = 0x200;
	regs.x.bx = 0;
	regs.x.dx = 0;
	int86(0x10, &regs, &regs);
}


void restcpos() {
	union REGS	regs;

	regs.x.ax = 0x200;
	regs.x.bx = 0;
	regs.x.dx = Curpos;
	int86(0x10, &regs, &regs);
}


void hplx_zoom(int enable) {
	union REGS	regs;

	if (!C_zoom_mode) return;
	regs.x.ax = 0xD000 | (enable? 0x81: 0x03);
	int86(0x10, &regs, &regs);
	if (enable)
		savecpos();
	else
		restcpos();
}
