#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <dos.h>
#include <dir.h>
#include "keys.h"
#include "vio.h"
#include "dirwin.h"
#include "util.h"
#include "conf.h"
#include "xpfv.h"
#include "xmem.h"


#define MAXDIRS		1000


struct dirrc {
	char	*name;
	int	depth;
};


char	Database[] = "\\xptree.xp5";

char	T_Spaces[] =	"        ";
char	T_Contbr[] =	"       ";
char	T_Branch[] =	"\020";
char	T_Lastbr[] =	"\020";
char	Spaces[9];
char	Contbr[9];
char	Branch[9];
char	Lastbr[9];

struct dirrc	Dirs[MAXDIRS];
int		Ndir;
int		Depth;
char		Qname[NAMELEN+1];
int		Qx;

struct viobuf	*info();


void convpath(char *p) {
	for (; *p; p++) {
		if (*p == '\\')
			*p = '/';
		else if (isupper(*p))
			*p = tolower(*p);
	}
}


int scan(depth, indx, indy) {
	struct ffblk	fb;
	int	nomore;
	static char	path[PATHL];
	static int	ind = 0;
	static char	*indc = "/-\\|";

	if (Ndir >= MAXDIRS) {
		alert("Too many directories", C_msg_time);
		return(-1);
	}
	getcwd(path, PATHL);
	convpath(path);
	Dirs[Ndir].name = xsave(path+2);
	Dirs[Ndir++].depth = depth;
	if (depth > Depth) Depth = depth;

	nomore = findfirst("*.*", &fb, 0x3f);
	while (!nomore) {
		vio_move(indx, indy);
		vio_putc(indc[ind++]);
		if (ind > 3) ind = 0;
		if ((fb.ff_attrib & FA_DIREC) && *fb.ff_name != '.') {
			chdir(fb.ff_name);
			scan(depth+1, indx, indy);
			chdir("..");
		}
		nomore = findnext(&fb);
	}
	return(0);
}


int rescan() {
	static char	path[PATHL];
	struct viobuf	*bg;
	char	*msg = "chdir() failed";
	int	rc;

	getcwd(path, PATHL);
	Depth = Ndir = 0;
	if (chdir("\\") < 0) {
		alert(msg, C_msg_time);
		return(-1);
	}
	bg = info("Rescanning directory structure:  ", NULL);
	rc = scan(0, bg->dx-2, bg->y+1);
	vio_hide(bg);
	if (chdir(path+2) < 0) {
		alert(msg, C_msg_time);
		return(-1);
	}
	return(rc);
}


int writetree() {
	FILE	*ddb;
	int	i;

	if ((ddb = fopen(Database, "w")) == NULL) {
		alert("Cannot create directory database", C_msg_time);
		return(-1);
	}

	for (i=0; i<Ndir; i++) {
		if (fprintf(ddb, "%3d %s\n", Dirs[i].depth, Dirs[i].name)
		== EOF) {
			alert("Error writing directory database", C_msg_time);
			fclose(ddb);
			return(-1);
		}
	}

	fclose(ddb);
	_chmod(Database, 1, FA_HIDDEN);
	return(0);
}


int readtree() {
	FILE	*ddb;
	char	buf[PATHL+1];

	if ((ddb = fopen(Database, "r")) == NULL) return(-1);
	fgets(buf, PATHL, ddb);
	for (Depth = Ndir = 0; !feof(ddb) && Ndir < MAXDIRS; ) {
		buf[strlen(buf)-1] = 0;
		Dirs[Ndir].name = xsave(buf+4);
		Dirs[Ndir].depth = atoi(buf);
		if (Dirs[Ndir].depth > Depth) Depth = Dirs[Ndir].depth;
		Ndir++;
		fgets(buf, PATHL, ddb);
	}
	fclose(ddb);
	return(0);
}


void deltree() {
	int	i;

	for (i=0; i<Ndir; i++) xfree(Dirs[i].name);
}


void qcdframe(int dww, int my) {
	int	dx, dy;

	vio_color(C_dir_frame_c);
	vio_frame(0, 0, dww-1, vio_lines-1);
	vio_color(C_dir_text_c);
	vio_box(1, 1, dww-2, vio_lines-2);
	vio_color(C_inf_text_c);
	vio_win(dww, 0, vio_clipx-1, my-1);
	vio_color(C_mnu_frame_c);
	vio_frame(dww, my, vio_clipx-1, vio_lines-1);
	vio_color(C_mnu_text_c);
	vio_box(dww+1, my+1, vio_clipx-2, vio_lines-2);

	if (vio_clipx > 64) {
		dx = 3;
		dy = 2;
	}
	else {
		dx = 2;
		dy = 1;
	}

	vio_move(dww, my+2);
 	vio_puts("");
 	while (vio_x < vio_clipx-1) vio_puts("");
	vio_puts("");

	vio_move(dww+dx, dy+my+3); vio_puts("Cursor Keys = Move");

	vio_move(dww+dx, dy+my+4); vio_puts("ESC         = Exit");
	vio_move(dww+dx, dy+my+5); vio_puts("Return      = Chdir");
	vio_move(dww+dx, dy+my+6); vio_puts("Left/Right  = Next/Prev");
	vio_move(dww+dx, dy+my+7); vio_puts("^R          = Rescan");
	vio_move(dww+dx, dy+my+8); vio_puts("Others      = Find");
}


int islast(int n, int base) {
	int	i;

	for (i=n+1; i<Ndir; i++) {
		if (Dirs[i].depth == base) return(0);
		if (Dirs[i].depth < base) return(1);
	}
	return(1);
}


void showbranch(int dww, int n) {
	int	i;
	char	*p;

	vio_putc(' ');
	for (i=0; i<Dirs[n].depth-1; i++) {
		vio_puts(islast(n, i+1)? Spaces: Contbr);
	}
	if (Dirs[n].depth) {
		vio_puts(islast(n, Dirs[n].depth)? Lastbr: Branch);
	}
	p = strrchr(Dirs[n].name, '/')+1;
	vio_puts(*p? p: p-1);
	while (vio_x < dww-2) vio_putc(' ');
}


void qcdinfo(int dww, int my) {
	struct dfree	dfb;
	char	buf[20];
	long	total, free;
	int	i;
	int	dx, dy;

	dx = dww>40? 3: 2;
	dy = dww>40? 1: 0;
	vio_move(dww+dx, 1+dy);
	sprintf(buf, "%8d", Ndir);
	vio_color(C_inf_high_c);
	vio_puts(buf);
	vio_color(C_inf_text_c);
	vio_puts("  Directories");

	vio_move(dww+dx, 2+dy);
	sprintf(buf, "%8d", Depth);
	vio_color(C_inf_high_c);
	vio_puts(buf);
	vio_color(C_inf_text_c);
	vio_puts("  Levels");

	getdfree(0, &dfb);
	if ((int) dfb.df_sclus == -1) return;
 	total = (long) dfb.df_sclus * dfb.df_bsec * dfb.df_total / 1024;
 	free = (long) dfb.df_sclus * dfb.df_bsec * dfb.df_avail / 1024;

	sprintf(buf, "%8ld KB", total);
	vio_move(dww+dx, dy+4);
	vio_color(C_inf_high_c);
	vio_puts(buf);
	vio_color(C_inf_text_c);
	vio_puts("  Capacity");

	sprintf(buf, "%8ld KB", free);
	vio_move(dww+dx, dy+5);
	vio_color(C_inf_high_c);
	vio_puts(buf);
	vio_color(C_inf_text_c);
	vio_puts("  Available");

	vio_move(dww+dx, my+1);
	vio_color(C_mnu_text_c);
	vio_puts("QuickFind: ");
	vio_color(C_mnu_high_c);
	for (i=0; i<NAMELEN; i++) vio_putc('' & 0xff);
}


int qfind(int dww, int my, int k) {
	int	i;
	char	*p;

	if (dww < 40) dww--;
	vio_color(C_mnu_high_c);
	if (!k) {
		vio_move(dww+14, my+1);
		for (i=0; i<NAMELEN; i++) vio_putc('' & 0xff);
		Qx = 0;
		return(-1);
	}

	if (k == '\b') {
		if (!Qx) return(-1);
		vio_move(dww+14+--Qx, my+1);
		Qname[Qx] = 0;
		vio_putc('' & 0xff);
		return(-1);
	}

	if (Qx >= NAMELEN) return(-1);

	Qname[Qx++] = k;
	Qname[Qx] = 0;

	for (i=0; i<Ndir; i++) {
		p = strrchr(Dirs[i].name, '/')+1;
		if (!strncmp(p, Qname, Qx)) break;
	}
	if (i >= Ndir) {
		Qname[--Qx] = 0;
		return(-1);
	}

	vio_move(dww+14-1+Qx, my+1);
	vio_putc(k);

	return(i);
}


int nextmatch(int pos) {
	int	i;
	char	*p;

	for (i=0; i<Ndir; i++) {
		if (++pos >= Ndir) pos = 0;
		p = strrchr(Dirs[pos].name, '/')+1;
		if (!strncmp(p, Qname, Qx)) return(pos);
	}
	return(pos);
}


int prevmatch(int pos) {
	int	i;
	char	*p;

	for (i=0; i<Ndir; i++) {
		if (--pos < 0) pos = Ndir-1;
		p = strrchr(Dirs[pos].name, '/')+1;
		if (!strncmp(p, Qname, Qx)) return(pos);
	}
	return(pos);
}


void adjust(int dww) {
	int	max;

	max = (dww-14)/Depth;
	strcpy(Spaces, T_Spaces);
	strcpy(Contbr, T_Contbr);
	strcpy(Branch, T_Branch);
	strcpy(Lastbr, T_Lastbr);
	if (max > 8) return;
	Spaces[max] = 0; Spaces[max-1] = T_Spaces[7];
	Contbr[max] = 0; Contbr[max-1] = T_Contbr[7];
	Branch[max] = 0; Branch[max-1] = T_Branch[7];
	Lastbr[max] = 0; Lastbr[max-1] = T_Lastbr[7];
}


int viewtree(int home) {
	struct viobuf	*scr;
	int	top, curr, k, i, j, red;
	int	dww, my;
	int	inc;

	timerc(0);
	scr = vio_protect(0, 0, 79, 24);
	dww = vio_clipx-(vio_clipx>64? 30: 27);
	my = vio_lines/3+1;
	inc = (vio_lines-2)/2;
	qcdframe(dww, my);
	curr = home;
	if ((top = curr - inc) < 0) top = 0;
	Qx = 0;
	qcdinfo(dww, my);
	adjust(dww);
	for (red = 1;; ) {
		if (red) {
			for (i=top; i<top+vio_lines-2 && i<Ndir; i++) {
				vio_move(2, i-top+1);
				vio_color((i==home)? C_dir_tag_c:
				C_dir_text_c);
				showbranch(dww, i);
			}
			vio_color(C_dir_text_c);
			for (i-=top; i<vio_lines-2; i++) {
				vio_move(2, i+1);
				for (j=0; j<dww-4; j++) vio_putc(' ');
			}
			red = 0;
		}
		vio_color((curr == home)? C_dir_tagbar_c: C_dir_bar_c);
		vio_move(2, curr-top+1);
		showbranch(dww, curr);
		k = key();
		vio_color((curr == home)? C_dir_tag_c: C_dir_text_c);
		vio_move(2, curr-top+1);
		showbranch(dww, curr);
		switch (k) {
			case K_ESC:
				vio_hide(scr); timerc(1); return(-1);

			case K_CR:
				vio_hide(scr); timerc(1); return(curr);

			case K_DOWN:
				if (curr >= Ndir-1) break;
				if (++curr >= top+vio_lines-2) {
					top++; red = 1;
				}
				break;

			case K_UP:
				if (curr < 1) break;
				if (--curr < top) {
					top--; red = 1;
				}
				break;

			case K_PGDN:
				if ((curr += vio_lines-3) >= Ndir)
					curr = Ndir-1;
				if ((top += vio_lines-3) >= Ndir) {
					if ((top = curr-(vio_lines-4)) < 0)
						top = 0;
				}
				red = 1;
				break;

			case K_PGUP:
				if ((curr -= vio_lines-3) < 0) curr = 0;
				if ((top -= vio_lines-3) < 0) top = 0;
				red = 1;
				break;

			case '/':
			case K_HOME:
				curr = top = 0; red = 1; break;

			case K_END:
				curr = Ndir-1;
				if ((top = curr-(vio_lines-4)) < 0) top = 0;
				red = 1;
				break;
		
			case ctl('R'):
				deltree();
				rescan();
				writetree();
				adjust(dww);
				home = searchhome();
				red = 1;
				Qx = 0;
				qcdinfo(dww, my);
				if (curr >= Ndir) curr = top = 0;
				break;

			case K_KILL:
			case ctl('U'):
				qfind(dww, my, 0); break;

			case '\b':
				qfind(dww, my, '\b'); break;

			case K_LEFT:
			case K_RGHT:
			default:
				if (k == K_RGHT) {
					if (!Qx) break;
					i = nextmatch(curr);
				}
				else if (k == K_LEFT) {
					if (!Qx) break;
					i = prevmatch(curr);
				}
				else {
					if ((i = qfind(dww, my, k)) < 0)
						break;
				}
				curr = i;
				if (i < top || i > top+vio_lines-3) {
					if ((top = curr-inc) < 0)
						top = 0;
					red = 1;
				}
		}
	}
}


int searchdir(char *dir) {
	char	*p;
	int	i, l;

	if (!(*dir)) return(-1);
	if (!dir[1] && *dir == '/' || *dir == '\\') return(0);
	for (i=0; i<Ndir; i++) {
		p = strrchr(Dirs[i].name, '/')+1;
		if (!strcmp(p, dir)) return(i);
	}
	l = strlen(dir);
	for (i=0; i<Ndir; i++) {
		p = strrchr(Dirs[i].name, '/')+1;
		if (!strncmp(p, dir, l)) return(i);
	}
	return(-1);
}


int searchhome() {
	int	i, tries = 0;
	char	path[PATHL+1];

	for (;;) {
		getcwd(path, PATHL);
		convpath(path);
		for (i=0; i<Ndir; i++) {
			if (!strcmp(Dirs[i].name, path+2)) return(i);
		}
		if (tries++) {
			alert("Oops! Rescan failed. Giving up.", 0);
			return(0);
		}
		alert("CWD not found -- will rescan now", C_msg_time);
		deltree();
		rescan();
		writetree();
	}
}


int qcd(char *dir, int force) {
	int	n, home;
	char	path[PATHL];

	if (force || readtree() < 0) {
		rescan();
		writetree();
	}
	n = searchdir(dir);
	home = searchhome();
	for (;;) {
		if (n<0) n = viewtree(home);
		if (n >= 0) {
			strcpy(path, Dirs[n].name);
			if (chdir(path) < 0) {
				alert("chdir() failed -- will rescan now",
				C_msg_time);
				deltree();
				rescan();
				writetree();
				n = -1;
			}
			else {
				break;
			}
		}
		else {
			break;
		}
	}
	deltree();
	return(n<0? -1: 0);
}
