#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <xspawn.h>
#include <dos.h>
#include <dir.h>
#include "keys.h"
#include "vio.h"
#include "dirwin.h"
#include "util.h"
#include "conf.h"
#include "xpfv.h"
#include "xmem.h"


struct viobuf	*Pathbg = NULL;


void xprun(char *cmd) {
	static char	*argv[64], buf[128];
	char	*p;
	int	argc;

	strcpy(buf, cmd);
	p = strtok(buf, " ");
	argc = 0;
	while (p) {
		argv[argc++] = p;
		p = strtok(NULL, " ");
	}
	argv[argc] = NULL;
	if (!argc) return;

	if (builtin(argc, argv)) return;

	(C_swap? xsystem: system)(cmd);
}


void xp_hideall(struct dirwin *dw, struct dirwin *dest) {
	if (dest) dw_hide(dest);
	dw_hide(dw);
	if (C_fullpath) {
		vio_hide(Pathbg); Pathbg = NULL;
	}
	hplx_zoom(0);
}


void xp_showall(struct dirwin *dw, struct dirwin *dest, struct dirwin *this) {
	hplx_zoom(1);
	if (C_fullpath && !Pathbg)
		Pathbg = vio_protect(0, vio_lines-1, 79, vio_lines-1);
	dw_show(dw);
	if (dest) dw_show(dest);
	if (this == dw) dw_refresh(dw, 0);
}


void xp_update(struct dirwin *dw, struct dirwin *dest, struct dirwin *this) {
	if (this == dest) dw_reread(dw, 1);
	if (dest) dw_reread(dest, 1);
	if (this == dw) dw_reread(dw, 1);
}


void xp_execute(struct dirwin *dw, struct dirwin *dest, struct dirwin *this,
char *cmd) {
	extern struct dirwin	*Main, *Dest, *Curr;
	char	path[PATHL];

	xp_hideall(dw, dest);
	Main = dw;
	Dest = dest;
	Curr = this;
	xprun(cmd);
	vio_init();
	xp_showall(dw, dest, this);
	getcwd(path, PATHL);
	if (stricmp(path, this->dir->path)) {
		this->curr = this->top = 0;
	}
	xfree(this->dir->path);
	this->dir->path = xsave(path);
	xp_update(dw, dest, this);
}


void xp_screen(struct dirwin *dw, struct dirwin *dest, struct dirwin *this,
int clear, int pause) {
	xp_hideall(dw, dest);
	if (clear) {
		vio_color(C_dos_color);
		vio_clear();
		vio_sync();
	}
	else if (pause) {
		timerc(0);
		key();
		timerc(1);
	}
	xp_showall(dw, dest, this);
}


void chattr(struct dirent *de, int attr) {
	char	buf[80];

	if (de->attr & FA_DIREC) {
		sprintf(buf, "\"%s\" is a directory", de->name);
		alert(buf, C_msg_time);
		return;
	}
	if (_chmod(de->name, 1, attr & 0x3f) < 0) {
		sprintf(buf, "Could not change attributes: \"%s\"", de->name);
		alert(buf, C_msg_time);
		return;
	}
	de->attr = attr;
}


void xp_attr(struct dirwin *dw, struct dirwin *dest, struct dirwin *this) {
	extern struct viobuf	*info();
	char	buf[80];
	int	attr, dflt = 0, i, ntags;

	attr = dw->dir->entries[dw->curr].attr;
	if ((ntags = have_tags(dw)) != 0) {
		attr = 0;
	}
	else if (attr & FA_DIREC) {
		alert("Cannot change attributes of a directory", C_msg_time);
		return;
	}

	for (;;) {
		sprintf(buf, "[%c] #Read Only|[%c] #Hidden|[%c] #System|"
		"[%c] #Archive|    #OK",
		(attr & FA_RDONLY)? 'X': ' ',
		(attr & FA_HIDDEN)? 'X': ' ',
		(attr & FA_SYSTEM)? 'X': ' ',
		(attr & FA_ARCH)? 'X': ' ');
		i = menu(dflt, buf);
		switch(i) {
			case -1:return;
			case 0:	attr = (attr & FA_RDONLY)? attr&~FA_RDONLY:
				attr|FA_RDONLY; break;
			case 1:	attr = (attr & FA_HIDDEN)? attr&~FA_HIDDEN:
				attr|FA_HIDDEN; break;
			case 2:	attr = (attr & FA_SYSTEM)? attr&~FA_SYSTEM:
				attr|FA_SYSTEM; break;
			case 3:	attr = (attr & FA_ARCH)? attr&~FA_ARCH:
				attr|FA_ARCH; break;
			case 4:	break;
		}
		if (i==4) break;
		dflt = i;
	}

	if (ntags) {
		char	buf[100];
		struct viobuf	*vb;

		sprintf(buf, "#Yes, change %d tagged files|"
		"#No, cancel this operation", ntags);
		if (menu(0, buf)) return;
		for (i=0; i<dw->dir->nfile; i++)
			if (dw->dir->entries[i].attr & E_TAGGED) {
				vb = info("Processing \"%s\"",
				dw->dir->entries[i].name);
				chattr(&dw->dir->entries[i], attr);
				vio_hide(vb);
			}
		dw_list(dw);
	}
	else {
		chattr(&dw->dir->entries[dw->curr], attr);
	}

	xp_update(dw, dest, this);
}


void xp_edit(struct dirwin *dw, struct dirwin *dest, struct dirwin *this) {
	char	cmd[128], *editor;
	extern char	*expand();

	if (!C_editor) {
		alert("No editor configured", C_msg_time);
		return;
	}
	if (this->dir->entries[this->curr].attr & FA_DIREC) {
		alert("Cannot edit directories", C_msg_time);
		return;
	}

	if (!exists(editor = expand(C_editor, NULL),
	"The configured editor does not exist"))
		return;
	sprintf(cmd, "%s %s", editor,
	this->dir->entries[this->curr].name);
	xp_execute(dw, dest, this, cmd);
}


void xp_view(struct dirwin *dw) {
	if (dw->dir->entries[dw->curr].attr & FA_DIREC) {
		alert("File is a directory", C_msg_time); return;
	}
	if (dw->dir->entries[dw->curr].size == 0L) {
		alert("File is empty", C_msg_time); return;
	}

	xpfv(dw->dir->entries[dw->curr].name, &C_browse_mode);
}


void xp_rename(struct dirwin *dw) {
	char	fname[NAMELEN+1], *old;

	if (dw->dir->entries[dw->curr].attr & FA_DIREC) {
		alert("Cannot rename directories", C_msg_time);
		return;
	}
	strcpy(fname, (old = dw->dir->entries[dw->curr].name));
	if (vio_prompt(-1, -1, -1, C_pmt_text_c, C_pmt_high_c, C_pmt_frame_c,
	NAMELEN, fname, "Rename") < 0) return;
	if (!(*fname)) return;

	if (rename(old, fname) < 0) {
		alert("Rename failed", C_msg_time);
		return;
	}

	dw_reread(dw, 0);
	dw_goto(dw, fname);
	dw_list(dw);
}


void move_win(struct dirwin *dw, struct dirwin *dest, struct dirwin *this) {
	struct viobuf	*vb;
	int	k, x, y, xl, yl;

	x = this->x;
	y = this->y;
	xl = this->xl;
	yl = this->yl;
	for (;;) {
		vb = vio_protect(x, y, x+xl, y+yl);
		vio_color(C_winctl_c);
		vio_frame(x, y, x+xl, y+yl);
		k = key();
		vio_hide(vb);
		switch(k) {
			case K_LEFT:	if (x>0) --x; break;
			case K_UP:	if (y>0) --y; break;
			case K_RGHT:	if (x+xl < vio_clipx-1) ++x; break;
			case K_DOWN:	if (y+yl < vio_lines-1) ++y; break;
			case K_HOME:
			case K_WLFT:	x = 0; break;
			case K_END:
			case K_WRGT:	x = vio_clipx-1-xl; break;
			case K_PGUP:	y = 0; break;
			case K_PGDN:	y = vio_lines-1-yl; break;
			case K_ESC:	return;
			case K_CR:	xp_hideall(dw, dest);
					this->x = x; this->xl = xl;
					this->y = y; this->yl = yl;
					xp_showall(dw, dest, this);
					return;
		}
	}
}


void resize_win(struct dirwin *dw, struct dirwin *dest, struct dirwin *this) {
	struct viobuf	*vb;
	int	k, yl;

	yl = this->yl;
	for (;;) {
		vb = vio_protect(this->x, this->y, this->x+this->xl,
			this->y+yl);
		vio_color(C_winctl_c);
		vio_frame(this->x, this->y, this->x+this->xl, this->y+yl);
		k = key();
		vio_hide(vb);
		switch(k) {
			case K_DOWN:	if (this->y + yl < vio_lines-1) yl++;
					break;
			case K_UP:	if (this->y+3 < yl) yl--; break;
			case K_HOME:	yl = 3; break;
			case K_END:	yl = vio_lines-1-this->y; break;
			case K_ESC:	return;
			case K_CR:	xp_hideall(dw, dest);
					if ((this->yl = yl) <
					this->curr-this->top + 2)
						this->curr = this->top +
						this->yl-2;
					xp_showall(dw, dest, this);
					return;
		}
	}
}


void xp_winctl(struct dirwin **dw, struct dirwin **dest, struct dirwin **this)
{
	int	i;
	char	mb[80];

	sprintf(mb, "#Move Window|#Resize Window%s",
	(*dest)? "|Only #1 window": "");
	if ((i = menu(0, mb)) < 0) return;
	switch (i) {
		case 0:	move_win(*dw, *dest, *this); break;
		case 1:	resize_win(*dw, *dest, *this); break;
		case 2:	if (*this == *dest) *this = *dw;
			dw_close(*dest);
			chdp((*dw)->dir->path);
			*dest = NULL;
			dw_refresh(*this, 1);
			break;
	}
}


void xp_action(struct dirwin *dw, struct dirwin *dest, struct dirwin *this,
int edit) {
	char	*name;
	static char	cmd[130];
	int	i;

	name = this->dir->entries[this->curr].name;

	for (i=0; Actions[i].pattern; i++) {
		if (fnmatch(Actions[i].pattern, name)) {
			strcpy(cmd, Actions[i].action);
			expand(cmd, this);
			if (edit) {
				strcat(cmd, " ");
				Cmd_ext = this;
				if (vio_prompt(0, vio_lines-3, vio_clipx-2,
				C_pmt_text_c, C_pmt_text_c, C_pmt_frame_c, 128,
				cmd, "Command") < 0) return;
			}
			if (!(*cmd)) return;
			xp_execute(dw, dest, this, cmd);
			return;
		}
	}

	alert("No default action for this file type", C_msg_time);
}


void xpf_sort(struct dirwin *dw, struct dirwin *dest, struct dirwin *this) {
	int 	new;

	new = menu(C_sort_mtd, "#Do not sort|Sort by #Name|Sort by #Suffix");
	if (new < 0) return;
	C_sort_mtd = new;
	if (this == dest) dw_reread(dw, 1);
	if (dest) dw_reread(dest, 1);
	if (this == dw) dw_reread(dw, 1);
}


void xp_files(struct dirwin *dw, struct dirwin *dest, struct dirwin *this) {
	switch (menu(0, "#Copy files|#Move files|#Delete files|#Sort files|"
	"Ma#ke directory|#Remove directory")) {
		case 0:	if (this == dw)
				copy_move(dw, dest, 0);
			else
				copy_move(dest, dw, 0);
			break;
		case 1:	if (this == dw)
				copy_move(dw, dest, 1);
			else
				copy_move(dest, dw, 1);
			break;
		case 2:	do_delete(this); break;
		case 3:	xpf_sort(dw, dest, this); break;
		case 4:	do_mkdir(this); break;
		case 5:	do_rmtree(this); break;
	}
}


void xp_swap(struct dirwin *dw, struct dirwin *dest, struct dirwin *this) {
	int	t;
	char	s[NAMELEN+1];
	struct dir	*d;

	if (!dest) return;
	t = dw->top; dw->top = dest->top; dest->top = t;
	t = dw->curr; dw->curr = dest->curr; dest->curr = t;
	d = dw->dir; dw->dir = dest->dir; dest->dir = d;
	strcpy(s, dw->pattern);
	strcpy(dw->pattern, dest->pattern);
	strcpy(dest->pattern, s);
	xp_screen(dw, dest, this, 0, 0);
}


void xp_setup(struct dirwin *dw, struct dirwin *dest, struct dirwin *this) {
	static char	xpsetup[PATHL], setupcmd[129];
	static char	home[PATHL];

	sprintf(xpsetup, "%s\\xpsetup.exe", C_home);
	if (!exists(xpsetup, "Setup program not found")) return;
	sprintf(setupcmd, "%s --quickstartup", xpsetup);
	xp_execute(dw, dest, this, setupcmd);
	strcpy(home, C_home);
	reconfigure(home);
	xp_screen(dw, dest, this, 0, 0);
}


void xp_qcd(struct dirwin *dw, struct dirwin *dest, struct dirwin *this) {
	char	dirname[NAMELEN+1];
	char	path[PATHL+1];

	*dirname = 0;
	if (vio_prompt(-1, -1, -1, C_pmt_text_c, C_pmt_high_c, C_pmt_frame_c,
	NAMELEN, dirname, "QuickCD") < 0) return;
	if (qcd(dirname, 0) < 0) return;
	getcwd(path, PATHL);
	if (stricmp(path, this->dir->path)) {
		this->curr = this->top = 0;
	}
	xfree(this->dir->path);
	this->dir->path = xsave(path);
	xp_update(dw, dest, this);
}


void xp_qff(struct dirwin *dw, struct dirwin *dest, struct dirwin *this) {
	char	fname[NAMELEN+4];
	char	path[PATHL+1];
	char	*where;
	extern char	*qff();

	*fname = 0;
	if (vio_prompt(-1, -1, -1, C_pmt_text_c, C_pmt_high_c, C_pmt_frame_c,
	NAMELEN, fname, "QuickFF") < 0) return;
	if (!(*fname)) return;
	strcat(fname, (strrchr(fname, '.'))? "*": "*.*");
	if ((where = qff(fname, 0)) == NULL) return;
	getcwd(path, PATHL);
	if (stricmp(path, this->dir->path)) {
		this->curr = this->top = 0;
	}
	xfree(this->dir->path);
	this->dir->path = xsave(path);
	xp_update(dw, dest, this);
	dw_goto(this, where);
}


int xp_fkey(struct dirwin *dw, struct dirwin *dest, struct dirwin *this,
int k) {
	char	buf[128];

	switch (k) {
		case K_F1: case K_F2: case K_F3: case K_F4: case K_F5:
		case K_F6: case K_F7: case K_F8: case K_F9: case K_F10:
			break;
		default:
			return(0);
	}

	if (!F_keys[k-K_F1]) {
		sprintf(buf, "Key F%d is not bound", k-K_F1+1);
		alert(buf, C_msg_time);
		return(1);
	}

	strcpy(buf, F_keys[k-K_F1]);
	expand(buf, this);
	xp_execute(dw, dest, this, buf);

	return(1);
}


int xp_command(struct dirwin **this, struct dirwin **dw, struct dirwin **dest,
int k) {
	switch(k) {
		case ctl('A'):	xp_attr(*dw, *dest, *this); break;
		case ctl('E'):	xp_edit(*dw, *dest, *this); break;
		case ctl('F'):	xp_files(*dw, *dest, *this);
				xp_update(*dw, *dest, *this); break;
		case ctl('K'):	xp_qff(*dw, *dest, *this); break;
		case ctl('L'):	xp_screen(*dw, *dest, *this, 1, 0); break;
		case ctl('N'):	xp_rename(*this); break;
		case ctl('O'):	xp_setup(*dw, *dest, *this); break;
		case ctl('Q'):	xp_qcd(*dw, *dest, *this); break;
		case ctl('S'):	xp_swap(*dw, *dest, *this); break;
		case ctl('U'):	xp_screen(*dw, *dest, *this, 0, 1); break;
		case ctl('V'):	xp_view(*this); break;
		case ctl('Z'):	xp_winctl(dw, dest, this); break;
		case '!':	diskfree(); break;
		case K_LF:	xp_action(*dw, *dest, *this, 1); break;
		case K_RGHT:	if (!C_arrow_nav) return(0);
		case K_CR:	xp_action(*dw, *dest, *this, 0); break;
		default:	if (xp_fkey(*dw, *dest, *this, k)) break;
				return(0);
	}
	return(1);
}
